// Author: Torsten Tränkner
// version: 0.3 2017-04-22

#include <getopt.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <math.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>

// used for the timestamp
#include <time.h>
#include <sys/time.h>

// used for the check of the noise level
#include <map>
#include <iostream>
#include <algorithm>
#include <vector>
#include <iterator>
#include <sstream>

#include "ook.h"
#include "rtl.h"

using namespace std;

// remote control from Pollin, 6 19, combined should be ~25 samples
const int POLLIN_COMBINED_MIN =  22;
const int POLLIN_COMBINED_MAX =  28;
const int POLLIN_SHORT_MIN =  4;
// bad signal at falling edge
const int POLLIN_SHORT_MAX = 12;
const int POLLIN_LONG_MIN  = 13;
const int POLLIN_LONG_MAX  = 22;

// remote control from Conrad, 11 25, combined should be ~34 samples
const int CONRAD_COMBINED_MIN = 30;
const int CONRAD_COMBINED_MAX = 40;
const int CONRAD_SHORT_MIN =  8;
const int CONRAD_SHORT_MAX = 14;
const int CONRAD_LONG_MIN  = 22;
const int CONRAD_LONG_MAX  = 28;

// switch from smart wares, short pause 7, long pause 27
const int SMARTWARES_SHORT_MIN =  4;
const int SMARTWARES_SHORT_MAX = 10;
const int SMARTWARES_LONG_MIN  = 24;
const int SMARTWARES_LONG_MAX  = 30;

// motion sensor from Multi Kon Trade 31 + 10, combined should be ~41 samples
const int MULTIKONTRADE_COMBINED_MIN = 38;
const int MULTIKONTRADE_COMBINED_MAX = 44;
const int MULTIKONTRADE_LONG_MIN  = 28;
const int MULTIKONTRADE_LONG_MAX  = 34;
const int MULTIKONTRADE_SHORT_MIN = 7;
const int MULTIKONTRADE_SHORT_MAX = 13;

// Pollin     has ~190 samples between frames
// Conrad     has ~136 samples between frames
// Smartwares has ~216 samples between frames
const int MINIMUM_SAMPLES_BETWEEN_FRAMES = 100;

// minimum number of samples at a low period to decide that the frame is finished
const int MINIMUM_LOW_LENGTH_NEW_REPETITIONS = 400;

// minimum number of samples at a low period to decide that the frame burst is finished (8 frames)
const int MINIMUM_LOW_LENGTH_RESET_BURST = 10000;

// usually 8 frames per burst, Multi Kon Trade uses 15 frames
int maximum_number_of_frames_per_burst = 8;

// number of frame repetition to decide that the frame is valid
// one repetition should be enough - these are two frames
const int NUMBER_OF_FRAME_REPETITIONS_TO_BE_SURE = 1;

// a frame contains at least 100 samples - this is used to ignore random peaks in the noise
const int MINIMUM_FRAME_LENGTH = 100;

// number of samples to ignore in case a high amplitude level contains spikes
//     spike
//   __| _____
// _|  \/     |__
// replaces a low pass filter
const int MINIMUM_PAUSE_LENGTH = 5;

// at least MINIMUM_PULSES_PER_FRAME pulses should be received per frame
const int MINIMUM_PULSES_PER_FRAME = 20;

// factor for threshold to peak of the amplitude
//  PEAK_THRESHOLD_FACTOR has to be low because of Pollin remote
const float PEAK_THRESHOLD_FACTOR = 0.1;

// verbose output == 1, see command line options
static int verbose = 0;

// 433.92 MHz
static uint32_t centerFrequency = 433920000;

// static uint32_t sampleRate = 256000;

// a high number of samples per second decreases the delay between rtlsdr_read_async / debugHandler
static uint32_t sampleRate = 1000000;

// change the threshold according to your environment
static float amplitude_threshold = 0.0001;

// threshold for 0 and 1 bits of the last frame
// this is used to check for silence
static float last_frame_threshold = 0;

// check for low amplitude level (silence) after received frames
static int silence_counter = 0;
static int long_silence_counter = 0;


// number of repetitions for the noise check
const int NOISE_CHECK_REPETITIONS = 2;

// factor for minimum threshold (adaptive threshold * MINIMUM_NOISE_THRESHOLD_FACTOR)
const int MINIMUM_NOISE_THRESHOLD_FACTOR = 5;
// const int MINIMUM_NOISE_THRESHOLD_FACTOR = 10;

// initialize the noise level at start up
static int initialize_noise_level = 0;

// get the noise level
static int noise_amplitude_counter = 0;

// use 10000 samples to check the noise level
const int MAXIMUM_NOISE_COUNTER = 10000;

// at least MINIMUM_NOISE_OCCURRENCE samples of MAXIMUM_NOISE_COUNTER of a certain value indicate noise
const int MINIMUM_NOISE_OCCURRENCE = 100;

// adapt the threshold for noise to the current environment (at start up)
static float adaptive_noise_threshold = 0;

// minimum noise threshold is MINIMUM_NOISE_THRESHOLD_FACTOR * adaptive_noise_threshold
static float minimum_noise_threshold = 0;
static float original_minimum_noise_threshold = 0;

// previous noise is stored to check a deviation that is too high
static float previous_noise_threshold = -1;

// count the number of occurrences of the amplitudes
static std::map<float, unsigned int> amplitude_occurrences;


// at least 3 samples need to be above the minimum amplitude threshold
const int MINIMUM_HIGH_LEVEL = 3;

// array to store the values since frame start
static float last_frame_values[5000];

// array to store the length of the amplitude pulse
static unsigned int frame_pulse_length_values[1000];

// count the repetitions of the frame
static int frame_repetitions = 0;

// frame counter belongs to frame_pulse_length_values
static int frame_counter = 0;

// finished frame counter counts the number of "complete" frames - even invalid frames
static int finished_frame_counter = 0;

// count the number of samples below the minimum noise threshold
static int frame_low_counter = 0;

// used for debugging and writing to file
static unsigned int static_counter = 0;

// true if the frame started
static bool frame_started = false;

// save the highest amplitude value (peak)
static float highest_peak = 0;
static float highest_peak_counter = 0;

// the last received frame
static string last_frame = "";

// the last received frame that seems to be valid
static string last_sure_frame = "";

// the last received frame from a burst of 8 frames that seems to be valid
static string last_sure_frame_burst = "";


// file name to store values for gnuplot
char *dump_filename = 0;
static FILE *dump_file_filtered = 0;
static FILE *dump_file_recorded = 0;

// used by signal handlers to stop cleanly.
static struct rtldev *rtlToStop = 0;

static void showHelp( FILE *f )
{
  fprintf(f,
    "Usage: ookd [-h] [-?] [-v] [-f frequency] [-d dump_filename]\n"
    "  -h | -? | --help                      display usage and exit\n"
    "  -v | --verbose                        verbose logging\n"
    "  -f nnnn | --frequency nnn             set center frequency, default 433920000\n"
    "  -d filename                           filename to dump raw data\n"
    "               PLEASE BE CAREFUL WITH THE OPTION -d AS IT\n"
    "               WILL WRITE A LOT OF DATA INTO A FILE\n"
    "               USE IT FOR DEBUGGING ONLY\n"
    "  -t threshold                          amplitude threshold / 10000\n"
  );
}

inline string get_time_stamp() {
  time_t rawtime;
  struct tm *timeinfo;
  char time_string[80];

  struct timeval tp;
  gettimeofday(&tp, NULL);
  long int milliseconds = tp.tv_usec / 1000;

  time(&rawtime);
  timeinfo = localtime(&rawtime);

  strftime(time_string, 80, "%F %T.000 ", timeinfo);

  // add the milliseconds to the time_string
  time_string[20] = ((milliseconds / 100) % 10) + 0x30;
  time_string[21] = ((milliseconds / 10) % 10) + 0x30;
  time_string[22] = (milliseconds % 10) + 0x30;

  return string(time_string);
}

static inline void reset_variables() {
  frame_counter = 0;
  frame_low_counter = 0;
  frame_started = false;
  highest_peak = 0;
  highest_peak_counter = 0;
  minimum_noise_threshold = original_minimum_noise_threshold;
}

static inline void evaluate_pulses(int pulse_counts) {

  // avoid frames that are too short (random noise)
  if (pulse_counts < MINIMUM_PULSES_PER_FRAME) {
    return;
  }

  if (verbose == 1) {
    cerr << "evaluate_pulses: " << pulse_counts << endl;
    for (int index = 0; index < pulse_counts; index++) {
      cerr << frame_pulse_length_values[index] << ",";
      if (index % 2 == 1) {
        cerr << " ";
      }
    }
    cerr << endl;

  }

  stringstream mystream;
  bool recognition_sure = true;

  int combined_first_bit  = frame_pulse_length_values[0] + frame_pulse_length_values[1];
  int combined_second_bit = frame_pulse_length_values[2] + frame_pulse_length_values[3];

  // make an educated guess what protocol this is

  // -------------- POLLIN --------------
  if ((combined_first_bit  > POLLIN_COMBINED_MIN) and (combined_first_bit  < POLLIN_COMBINED_MAX) and
      (combined_second_bit > POLLIN_COMBINED_MIN) and (combined_second_bit < POLLIN_COMBINED_MAX))
  {

    // POLLIN == 6 high + 19 low
    mystream << "POLLIN: ";

    for (int index = 0; index < (pulse_counts - 1); index = index + 2) {
      int combined = frame_pulse_length_values[index] + frame_pulse_length_values[index + 1];
      if ( (combined > POLLIN_COMBINED_MIN) and (combined < POLLIN_COMBINED_MAX) ) {
        if (frame_pulse_length_values[index] < frame_pulse_length_values[index + 1]) {
          mystream << "1";
        } else {
          mystream << "0";
        }
      } else {
        // unknown combination
        mystream << "U";
        recognition_sure = false;
      }
    }

    if (mystream.str().length() != 32) {
      recognition_sure = false;
    }

  // -------------- CONRAD --------------
  } else if ((combined_first_bit  > CONRAD_COMBINED_MIN) and (combined_first_bit  < CONRAD_COMBINED_MAX) and
             (combined_second_bit > CONRAD_COMBINED_MIN) and (combined_second_bit < CONRAD_COMBINED_MAX))
  {

    // CONRAD == 25 high + 11 low

    mystream << "CONRAD: ";
    for (int index = 0; index < (pulse_counts - 1); index = index + 2) {
      if (index == 16) {
        mystream << " ";
      }
      int combined = frame_pulse_length_values[index] + frame_pulse_length_values[index + 1];
      if ( (combined > CONRAD_COMBINED_MIN) and (combined < CONRAD_COMBINED_MAX) ) {
        if (frame_pulse_length_values[index] < frame_pulse_length_values[index + 1]) {
          mystream << "1";
        } else {
          mystream << "0";
        }
      } else {
        // unknown combination
        mystream << "U";
        recognition_sure = false;
      }
    }

    if (mystream.str().length() != 41) {
      recognition_sure = false;
    }

  // -------------- SMARTWARES --------------
  } else if ((frame_pulse_length_values[0] > 2)  and (frame_pulse_length_values[0] < 7) and
             (frame_pulse_length_values[1] > 49) and (frame_pulse_length_values[1] < 55))
  {

    // SMARTWARES == short pause 7, long pause 27
    mystream << "SMARTWARES: ";

    for (int index = 3; index < (pulse_counts - 1); index = index + 2) {
      if (index == 111) {
        mystream << " ";
      }
      if ((frame_pulse_length_values[index] > SMARTWARES_SHORT_MIN) and (frame_pulse_length_values[index] < SMARTWARES_SHORT_MAX))
      {
        mystream << "1";
      }
      else if ((frame_pulse_length_values[index] > SMARTWARES_LONG_MIN) and (frame_pulse_length_values[index] < SMARTWARES_LONG_MAX))
      {
        mystream << "0";
      }
      else {
        mystream << "U";
        recognition_sure = false;
      }
    }

    // cout << "length: " << mystream.str().length() << endl;
    if (mystream.str().length() != 77) {
      recognition_sure = false;
    }

  // -------------- MULTI KON TRADE --------------
  } else if ((combined_first_bit  > MULTIKONTRADE_COMBINED_MIN) and (combined_first_bit  < MULTIKONTRADE_COMBINED_MAX) and
             (combined_second_bit > MULTIKONTRADE_COMBINED_MIN) and (combined_second_bit < MULTIKONTRADE_COMBINED_MAX))
  {

    // MULTIKONTRADE == 31 high + 10 low, vice versa

    mystream << "MULTIKONTRADE: ";
    for (int index = 0; index < (pulse_counts - 1); index = index + 2) {
      int combined = frame_pulse_length_values[index] + frame_pulse_length_values[index + 1];
      if ( (combined > MULTIKONTRADE_COMBINED_MIN) and (combined < MULTIKONTRADE_COMBINED_MAX) ) {
        if (frame_pulse_length_values[index] < frame_pulse_length_values[index + 1]) {
          mystream << "1";
        } else {
          mystream << "0";
        }
      } else {
        // unknown combination
        mystream << "U";
        recognition_sure = false;
      }
    }

    if (recognition_sure == true) {
      if (last_sure_frame == "") {
        maximum_number_of_frames_per_burst = 15;
      }
    }

  // -------------- unknown frame - print values --------------
  } else {

    recognition_sure = false;

    mystream << "unknown frame: " << pulse_counts << " static_counter: " << static_counter << endl;

    for (int index = 0; index < pulse_counts; index++) {
      mystream << frame_pulse_length_values[index] << ",";
      if (index % 2 == 1) {
        mystream << " ";
      }
    }

  }

  string display_frame = mystream.str();

  string recognition_sure_string = "";

  // if the received frame seems to be valid
  if (recognition_sure) {
    recognition_sure_string = " sure";

    if (verbose == 1) {
      cerr << "  last_frame: " << last_frame << endl;
      cerr << "  last_sure_frame: " << last_sure_frame << endl;
    }

    // first new frame since pause
    if (last_frame != "") {

      // if last frame is the same as the current frame
      if (last_frame == mystream.str()) {

        // increase repetition counter
        frame_repetitions++;

        // repetition within burst
        if (frame_repetitions >= NUMBER_OF_FRAME_REPETITIONS_TO_BE_SURE) {
          recognition_sure_string = " repetition sure";

          // set last sure frame
          last_sure_frame = last_frame;

          // start counting finished frames to find the end of the burst
          if (finished_frame_counter == 0) {
            finished_frame_counter = NUMBER_OF_FRAME_REPETITIONS_TO_BE_SURE;
          }
        }

      } else {

        // if no repetition of sure use this frame
        if (last_sure_frame == "") {
          // save current frame for the next check
          last_frame = mystream.str();

          // reset variables
          frame_repetitions = 0;
          finished_frame_counter = 0;

          if (verbose == 1) {
            cerr << "  changed last_frame: " << last_frame << endl;
          }
        } else {
          if (verbose == 1) {
            cerr << "  did not change last_frame, last_sure_frame: " << last_sure_frame << endl;
          }
        }
      }
    } else {

      // save current frame for the next check
      last_frame = mystream.str();

      if (verbose == 1) {
        cerr << "  stored first last_frame: " << last_frame << endl;
      }

    }
  }

  if (finished_frame_counter > 0) {
    finished_frame_counter++;

    // additional check in case that there is no pause between the bursts of 8 frames
    // > 7 repetitions
    if (finished_frame_counter > maximum_number_of_frames_per_burst - 1) {

      // check if the same key is still pressed
      if (last_sure_frame_burst == last_frame) {
        recognition_sure_string = " repetition burst sure";

      // another key is pressed - so no burst repetition
      } else {
        recognition_sure_string = " burst sure";
      }

      last_sure_frame_burst = last_sure_frame;
      display_frame = last_sure_frame;

      // reset the variables
      finished_frame_counter = 0;
      maximum_number_of_frames_per_burst = 8;
      frame_repetitions = 0;
      last_frame = "";
      last_sure_frame = "";
      silence_counter = 0;

    }

  }

  // print received frame
  cout << get_time_stamp() << display_frame << recognition_sure_string << " " << frame_repetitions << ", " << finished_frame_counter << endl;
  if (verbose == 1) {
    cerr << get_time_stamp() << display_frame << recognition_sure_string << " " << frame_repetitions << ", " << finished_frame_counter << endl;
  }

}

static inline void recognize_bits(float value) {

  // reset frame - in case it is too long
  if (frame_counter > 4900) {
    reset_variables();
  }

  if (frame_repetitions > 0) {

    if (long_silence_counter == 0) {
      long_silence_counter = 1;
    }

    if (value < last_frame_threshold) {

      silence_counter++;

      // check for end of one burst with 8 frames and silence afterwards
      if (silence_counter > MINIMUM_LOW_LENGTH_NEW_REPETITIONS) {
        if (last_sure_frame_burst == last_frame) {
          cout << get_time_stamp() << last_frame << " repetition burst sure (after timeout)" << endl;
        } else {
          cout << get_time_stamp() << last_sure_frame << " burst sure (after timeout)" << endl << endl;
        }
        last_sure_frame_burst = last_sure_frame;

        // reset variables
        silence_counter = 0;
        frame_repetitions = 0;
        finished_frame_counter = 0;
        maximum_number_of_frames_per_burst = 8;
        last_frame = "";
        last_sure_frame = "";
      }

    } else {
      silence_counter = 0;
    }
  }

  // check longer period of low amplitude level
  if (long_silence_counter > 0) {
    if (value < last_frame_threshold) {
      long_silence_counter++;

      // reset values after the several bursts are finished (no more repetitions)
      if (long_silence_counter > MINIMUM_LOW_LENGTH_RESET_BURST) {

        // reset variables
        long_silence_counter = 0;
        frame_repetitions = 0;
        finished_frame_counter = 0;
        maximum_number_of_frames_per_burst = 8;

        // reset frame from last burst !!!
        last_sure_frame_burst = "";
        last_sure_frame = "";
        last_frame = "";
      }
    } else {
      long_silence_counter = 0;
    }
  }

  // high
  if (value > minimum_noise_threshold) {
    if (frame_started == false) {
      frame_started = true;
    } else {
      if (value > highest_peak) {
        // calculate something like an "average" of peak values
        highest_peak = (value + highest_peak * highest_peak_counter) / (highest_peak_counter + 1);
        highest_peak_counter++;
        float part = highest_peak * PEAK_THRESHOLD_FACTOR;
        if (part > minimum_noise_threshold) {
          minimum_noise_threshold = part;
        }
      }
    }
    last_frame_values[frame_counter] = value;
    frame_counter++;
    frame_low_counter = 0;

    // cout << "value: " << value << endl;

  // low
  } else {
    if (frame_started == true) {
      last_frame_values[frame_counter] = value;
      frame_counter++;

      if (frame_counter < MINIMUM_HIGH_LEVEL) {

        reset_variables();

      } else {
        frame_low_counter++;
        if (frame_low_counter > MINIMUM_SAMPLES_BETWEEN_FRAMES) {
          if (frame_counter > MINIMUM_FRAME_LENGTH) {

            // set a threshold for 0 and 1 bits
            float frame_threshold = highest_peak * PEAK_THRESHOLD_FACTOR;
            if (frame_threshold < minimum_noise_threshold) {
              frame_threshold = minimum_noise_threshold;
            }

            last_frame_threshold = frame_threshold;

            if (verbose == 1) {
              cerr << "frame_counter: " << frame_counter << endl;
              cerr << "highest_peak: " << highest_peak << ", frame_threshold: " << frame_threshold << endl << endl;
            }

            int start_high = -1;
            int start_low = -1;
            int frame_pulse_length_counter = 0;

            for (int index = 0; index < frame_counter; index++) {

              // high
              if (last_frame_values[index] > frame_threshold) {
                if (start_high == -1) {
                  start_high = index;
                  if (start_low != -1) {
                    int pause_length = start_high - start_low;

                    // check if pause was just a peak
                    if (pause_length < MINIMUM_PAUSE_LENGTH) {
                      frame_pulse_length_counter--;
                      start_high = start_low - frame_pulse_length_values[frame_pulse_length_counter];
                      if (verbose == 1) {
                        cerr << "found peak at: " << start_high << " - " << frame_pulse_length_counter << endl;
                      }
                    // set low value
                    } else {
                      frame_pulse_length_values[frame_pulse_length_counter] = pause_length;
                      frame_pulse_length_counter++;
                    }
                    start_low = -1;
                  }
                }

              // low
              } else {
                if (start_high != -1) {
                  start_low = index;
                  int pulse_length = start_low - start_high;
                  // set high value
                  frame_pulse_length_values[frame_pulse_length_counter] = pulse_length;
                  frame_pulse_length_counter++;
                  start_high = -1;
                }
              }

              // see frame_pulse_length_values[]
              if (frame_pulse_length_counter > 900) {
                cerr << "pulse array is too short" << endl;
                exit(1);
              }
            }

            // frame finished
            evaluate_pulses(frame_pulse_length_counter);
          }

          reset_variables();
        }
      }
    }
  }
}

static inline void set_noise_level(float value) {

  amplitude_occurrences[value]++;
  noise_amplitude_counter++;

  // get noise level
  if ( noise_amplitude_counter >= MAXIMUM_NOISE_COUNTER ) {

    for(map<float, unsigned int>::iterator iter = amplitude_occurrences.begin(); iter != amplitude_occurrences.end(); ++iter) {

      if (verbose == 1) {
        float key =  iter->first;
        unsigned int value = iter->second;
        cerr << "Key: " << key << " value: " << value << endl;
      }

      if (iter->second > MINIMUM_NOISE_OCCURRENCE) {
        // adapt the noise threshold
        if (adaptive_noise_threshold < iter->first) {
          adaptive_noise_threshold = iter->first;
        }
      }

    }

    if (previous_noise_threshold == -1) {
      previous_noise_threshold = adaptive_noise_threshold;
    } else if (adaptive_noise_threshold > (previous_noise_threshold * 2) ) {
      adaptive_noise_threshold = previous_noise_threshold;
    }

    minimum_noise_threshold = adaptive_noise_threshold * MINIMUM_NOISE_THRESHOLD_FACTOR;
    original_minimum_noise_threshold = minimum_noise_threshold;

    cout << endl << get_time_stamp() << "adaptive_noise_threshold: " << adaptive_noise_threshold
         << ", minimum_noise_threshold: " << minimum_noise_threshold << endl << endl;

    cerr << endl << get_time_stamp() << "adaptive_noise_threshold: " << adaptive_noise_threshold
         << ", minimum_noise_threshold: " << minimum_noise_threshold << endl << endl;

    amplitude_occurrences.clear();
    noise_amplitude_counter = 0;

    previous_noise_threshold = adaptive_noise_threshold;
    initialize_noise_level++;
  }

}


static void debugPulses( const unsigned char *data, uint32_t len, uint8_t bins, const float alpha)
{
  // length is 262144 for my rtl sdr configuration
  if (verbose == 1) {
    // cerr << get_time_stamp() << "=== length: " << len << endl;
  }

  int i = 0;

  // take every 100th value to reduce CPU load
  for (i = 0; i < len; i += 100) {

    // create a float from the byte
    float I = (data[i]-128)/128.0;
    float Q = (data[i+1]-128)/128.0;

    // create a magnitude squared
    float powerSquared = I*I + Q*Q;

    static_counter++;

    if (dump_file_filtered != 0) {
      fprintf(dump_file_recorded, "%d %.5f\n", static_counter, powerSquared);
    }

    // first initialize the noise level
    if (initialize_noise_level < NOISE_CHECK_REPETITIONS) {
      set_noise_level(powerSquared);

    // then recognise smart home devices
    } else {
      recognize_bits(powerSquared);
    }

  }

}

static void debugHandler(const unsigned char *data, uint32_t len, void *ctx, struct rtldev *rtl)
{
  debugPulses( data, len, 16, 0.2);
}

static void exitNicely(int signum)
{
  if ( rtlToStop) {
    rtlStop( rtlToStop);
    rtlToStop = 0;
  } else {
    // we are stuck on something else
    exit(0);
  }
}

int main( int argc, char **argv)
{
  // check the options
  while (1 == 1) {
    int optionIndex = 0;

    static struct option options[] = {
      { "verbose", no_argument, 0, 'v' },
      { "help",    no_argument, 0, 'h' },
      { "frequency", required_argument, 0, 'f' },
      { "dump_filename", required_argument, 0, 'd'},
      { "multicast-port", required_argument, 0, 'p' },
      { "multicast-interface", required_argument, 0, 'i' },
      { "threshold", required_argument, 0, 't' },
      { 0,0,0,0}
    };

    int c = getopt_long( argc, argv, "vh?f:d:p:i:t:", options, &optionIndex );

    if ( c == -1 ) {
      break;
    }

    switch(c) {
      case 'h':
      case '?':
        showHelp(stdout);
        return 0;
      case 'v':
        verbose = 1;
        break;
      case 'f':
        // set frequency to optarg
        {
          unsigned f = atoi( optarg);
          if (f==0) {
            fprintf(stderr,"Bad frequency: %s\n", optarg);
            exit(1);
          }
          centerFrequency = f;
        }
        break;
      case 't':
        // set threshold to optarg
        {
          unsigned threshold = atoi( optarg);
          amplitude_threshold = threshold / 10000.0f;
        }
        break;
      case 'd':
        // dump file
        {
          dump_filename = optarg;

          char dump_filename_filtered[1000];
          snprintf(dump_filename_filtered, sizeof dump_filename_filtered, "%s%s", dump_filename, "_filtered.txt");

          char dump_filename_recorded[1000];
          snprintf(dump_filename_recorded, sizeof dump_filename_recorded, "%s%s", dump_filename, "_recorded.txt");

          dump_file_filtered = fopen ( dump_filename_filtered, "w");
          if (dump_file_filtered == NULL)
          {
            fprintf(stderr, "file open failed");
            exit(1);
          }

          dump_file_recorded = fopen ( dump_filename_recorded, "w");
          if (dump_file_recorded == NULL)
          {
            fprintf(stderr, "file open failed");
            exit(1);
          }

        }
        break;
      default:
        fprintf(stderr,"Illegal option\n");
        showHelp(stderr);
        exit(1);
    }
  }

  signal(SIGINT, exitNicely);

  struct rtldev *rtl = rtlOpen(NULL,0);

  if ( !rtl ) {
    fprintf(stderr,"Failed to open RTL SDR device\n");
    exit(1);
  }

  // 2 db
  // int gain = 20;
  int gain = 0;
  if ( rtlSetup( rtl, centerFrequency, sampleRate, gain) < 0) {
    fprintf(stderr,"Failed to setup RTL SDR for %u Hz %u samp/sec\n", centerFrequency, sampleRate);
  }

  // something must call rtlStop(rtl) to kill this, to this end we stash in a global, ick
  rtlToStop = rtl;
  if ( rtlRun( rtl, debugHandler, 0)) {
    fprintf(stderr, "Failed to run debugHandler\n");
  }

  rtlToStop = 0;
  rtlClose(rtl);

  return 0;
}
